UNIT UDNSClient;

INTERFACE

USES
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  WSockets, WinSock, StdCtrls, ComCtrls, ExtCtrls, Menus, Registry;

CONST

  NSListFile = 'nservers.txt';

  QueryID    : WORD = 0;   // Queryversion

  MaxTimeOut : WORD = 20;  // 20 Sec
  PosDiff    = 12;         // Bufferoffset

  DNSPort    = '53';

  Header_Option   : STRING[10] = #1#0#0#1#0#0#0#0#0#0; //  Headersection without VersionID

  QClass_Internet = #0#1;

  QType_A     = 1;
  QType_NS    = 2;
  QType_CNAME = 5;
  QType_SOA   = 6;
  QType_PTR   = 12;
  QType_MX    = 15;
  QType_ALL   = 255;

  QType_A_Str     = 'A';
  QType_NS_Str    = 'NS';
  QType_CNAME_Str = 'CNAME';
  QType_SOA_Str   = 'SOA';
  QType_PTR_Str   = 'PTR';
  QType_MX_Str    = 'MX';
  QType_ALL_Str   = 'ALL';


TYPE

  TLabelStr = STRING[255];

  TuLong    = RECORD
    CASE INTEGER OF
      0 : (
            Value : CARDINAL;
          );
      1 : (
            B : ARRAY[0..3] OF BYTE;
          );
  END;

  TuShort   = RECORD
    CASE INTEGER OF
      0 : (
            Value : WORD;
          );
      1 : (
            B : ARRAY[0..1] OF BYTE;
          );
  END;

  TBufferRec = RECORD
    Count   : WORD;
    ID      : WORD;
    CASE INTEGER OF
     0 : (
           Option  : ARRAY[0..1] OF CHAR;
           QDCount : ARRAY[0..1] OF CHAR;
           ANCount : ARRAY[0..1] OF CHAR;
           NSCount : ARRAY[0..1] OF CHAR;
           ARCount : ARRAY[0..1] OF CHAR;
           Data    : ARRAY[0..511] OF CHAR;
         );
     1 : (
           nOption  : WORD;
           nQDCount : WORD;
           nANCount : WORD;
           nNSCount : WORD;
           nARCount : WORD;
           nData    : ARRAY[0..511] OF BYTE;
         )
  END;

  TfrmMain = CLASS(TForm)
    edData: TEdit;
    btnQuery: TButton;
    StatusBar1: TStatusBar;
    cbType: TComboBox;
    memoLog: TMemo;
    Panel1: TPanel;
    Label1: TLabel;
    Label3: TLabel;
    cbHost: TComboBox;
    Label2: TLabel;
    mnuMain: TMainMenu;
    mnuFile: TMenuItem;
    mnuPreferences: TMenuItem;
    mnuAbout: TMenuItem;
    mnuClose: TMenuItem;
    mnuSep1: TMenuItem;
    mnuSaveLog: TMenuItem;
    dlgSaveLog: TSaveDialog;
    PROCEDURE edChange(Sender: TObject);
    PROCEDURE FormCreate(Sender: TObject);
    PROCEDURE FormDestroy(Sender: TObject);
    PROCEDURE btnQueryClick(Sender: TObject);
    procedure mnuAboutClick(Sender: TObject);
    procedure mnuCloseClick(Sender: TObject);
    procedure mnuSaveLogClick(Sender: TObject);
    procedure mnuPreferencesClick(Sender: TObject);
  PRIVATE
    UDPClient : TUDPClient;
    Buffer    : TBufferRec;
    TimeOut   : TTimer;
    Waiting   : BOOLEAN;

    PROCEDURE OnTimeOut(Sender: TObject);

    PROCEDURE UDPClientError(Sender: TObject; Error: integer; Msg: STRING);
    PROCEDURE UDPClientData(Sender: TObject; Socket: TSocket);
    PROCEDURE StringToLabel(Str : STRING; VAR Buf : TLabelStr);
    PROCEDURE LabelToString(Buf : POINTER; VAR Pos : WORD; VAR Str : String);
    FUNCTION  GetQType(qt : STRING): STRING;
  PUBLIC
    PROCEDURE TTY(Msg: STRING);
  END;

VAR
  frmMain: TfrmMain;


{--------------------------------------------------------}
IMPLEMENTATION

USES UAbout, UPreferences;

{$R *.DFM}

PROCEDURE TfrmMain.FormCreate(Sender: TObject);
VAR Reg : TRegistry;
BEGIN
  UDPClient := TUDPClient.Create(Self);
  WITH UDPClient DO BEGIN
    OnError := UDPClientError;
    OnData := UDPClientData;
  END;
  cbHost.Items.LoadFromFile(ExtractFileDir(Application.ExeName)+'\'+NSListFile);
  IF cbHost.Items.Count > 0 THEN cbHost.Text := cbHost.Items[0];
  {-}
  Reg := TRegistry.Create;
  TRY
    Reg.RootKey := HKey_Local_Machine;
    Reg.OpenKey('\SOFTWARE\NSinspector',TRUE);
    {-}
    IF NOT Reg.ValueExists('BackgroundColor') THEN Reg.WriteInteger('BackgroundColor',memoLog.Color);
    IF NOT Reg.ValueExists('Font') THEN Reg.WriteString('Font',memoLog.Font.Name);
    IF NOT Reg.ValueExists('FontColor') THEN Reg.WriteInteger('FontColor',memoLog.Font.Color);
    IF NOT Reg.ValueExists('FontSize') THEN Reg.WriteInteger('FontSize',memoLog.Font.Size);
    IF NOT Reg.ValueExists('FontStyleBold') THEN
      IF fsBold IN memoLog.Font.Style THEN Reg.WriteBool('FontStyleBold',TRUE)
                                      ELSE Reg.WriteBool('FontStyleBold',FALSE);
    IF NOT Reg.ValueExists('FontStyleItalic') THEN
      IF fsItalic IN memoLog.Font.Style THEN Reg.WriteBool('FontStyleItalic',TRUE)
                                        ELSE Reg.WriteBool('FontStyleItalic',FALSE);
    IF NOT Reg.ValueExists('FontStyleUnderline') THEN
      IF fsUnderline IN memoLog.Font.Style THEN Reg.WriteBool('FontStyleUnderline',TRUE)
                                           ELSE Reg.WriteBool('FontStyleUnderline',FALSE);
    IF NOT Reg.ValueExists('FontStyleStrikeout') THEN
      IF fsStrikeout IN memoLog.Font.Style THEN Reg.WriteBool('FontStyleStrikeout',TRUE)
                                           ELSE Reg.WriteBool('FontStyleStrikeout',FALSE);
    IF NOT Reg.ValueExists('Timeout') THEN Reg.WriteInteger('Timeout',MaxTimeOut);

    memoLog.Color      := Reg.ReadInteger('BackgroundColor');
    memoLog.Font.Name  := Reg.ReadString('Font');
    memoLog.Font.Color := Reg.ReadInteger('FontColor');
    memoLog.Font.Size  := Reg.ReadInteger('FontSize');
    memoLog.Font.Style := [];
    IF Reg.ReadBool('FontStyleBold') THEN memoLog.Font.Style := memoLog.Font.Style+[fsBold];
    IF Reg.ReadBool('FontStyleItalic') THEN memoLog.Font.Style := memoLog.Font.Style+[fsItalic];
    IF Reg.ReadBool('FontStyleUnderline') THEN memoLog.Font.Style := memoLog.Font.Style+[fsUnderline];
    IF Reg.ReadBool('FontStyleStrikeout') THEN memoLog.Font.Style := memoLog.Font.Style+[fsStrikeout];

    MaxTimeOut := Reg.ReadInteger('Timeout');

  FINALLY
    Reg.Free;
  END;
END;

PROCEDURE TfrmMain.FormDestroy(Sender: TObject);
VAR Reg : TRegistry;
BEGIN
  UDPClient.Free;
  cbHost.Items.SaveToFile(NSListFile);
  {-}
  Reg := TRegistry.Create;
  TRY
    Reg.RootKey := HKey_Local_Machine;
    Reg.OpenKey('\SOFTWARE\NSinspector',TRUE);
    {-}
    Reg.WriteInteger('BackgroundColor',memoLog.Color);
    Reg.WriteString('Font',memoLog.Font.Name);
    Reg.WriteInteger('FontColor',memoLog.Font.Color);
    Reg.WriteInteger('FontSize',memoLog.Font.Size);
    IF fsBold IN memoLog.Font.Style THEN Reg.WriteBool('FontStyleBold',TRUE)
                                    ELSE Reg.WriteBool('FontStyleBold',FALSE);
    IF fsItalic IN memoLog.Font.Style THEN Reg.WriteBool('FontStyleItalic',TRUE)
                                      ELSE Reg.WriteBool('FontStyleItalic',FALSE);
    IF fsUnderline IN memoLog.Font.Style THEN Reg.WriteBool('FontStyleUnderline',TRUE)
                                         ELSE Reg.WriteBool('FontStyleUnderline',FALSE);
    IF fsStrikeout IN memoLog.Font.Style THEN Reg.WriteBool('FontStyleStrikeout',TRUE)
                                         ELSE Reg.WriteBool('FontStyleStrikeout',FALSE);
    Reg.WriteInteger('Timeout',MaxTimeOut);
  FINALLY
    Reg.Free;
  END;
END;

PROCEDURE TfrmMain.mnuCloseClick(Sender: TObject);
BEGIN
  Application.Terminate;
END;

PROCEDURE TfrmMain.mnuAboutClick(Sender: TObject);
BEGIN
  frmAbout.ShowModal;
END;

PROCEDURE TfrmMain.mnuSaveLogClick(Sender: TObject);
BEGIN
  IF dlgSaveLog.Execute THEN memoLog.Lines.SaveToFile(dlgSaveLog.FileName);
END;

PROCEDURE TfrmMain.mnuPreferencesClick(Sender: TObject);
BEGIN
  frmPreferences.ShowModal;
END;

{-------------------------------------}

PROCEDURE TfrmMain.TTY(Msg: STRING);
BEGIN
  WITH memoLog.Lines DO BEGIN
    IF Count > 100 THEN
      Delete(0);
    Add(Msg);
  END;
END;

PROCEDURE TfrmMain.edChange(Sender: TObject);
BEGIN
  btnQuery.Enabled := (cbHost.Text <> '') AND (edData.Text <> '');
END;

PROCEDURE TfrmMain.UDPClientError(Sender: TObject; Error: INTEGER; Msg: STRING);
BEGIN
  MessageDlg(Msg, mtError, [mbOK], 0);
END;

PROCEDURE TfrmMain.OnTimeOut(Sender: TObject);
BEGIN
  TimeOut.Enabled := FALSE;
  TTY('Max Timeout exceeded ...');
  TTY('');
  Waiting := FALSE;
END;

{-------------------------------------}

PROCEDURE TfrmMain.StringToLabel(Str : STRING; VAR Buf : TLabelStr);
VAR SubLabel : STRING;
BEGIN
  Buf := '';
  WHILE (Length(Str) > 0) AND (Pos('.',Str) > 0) DO BEGIN
    SubLabel := Copy(Str,1,Pos('.',Str)-1);
    Delete(Str,1,Pos('.',Str));
    Buf := Buf+Chr(Length(SubLabel))+SubLabel;
  END;
  IF Length(Str) > 0 THEN Buf := Buf+Chr(Length(Str))+Str;
  Buf := Buf+#0;  // Rootlabel = #0
END;

PROCEDURE TfrmMain.LabelToString(Buf : POINTER; VAR Pos : WORD; VAR Str : String);
VAR SubLabel : PChar;
    i,j, Len : INTEGER;
BEGIN
  Str := '';
  SubLabel := Buf;
  i := Pos;
  Pos := 0;
  WHILE SubLabel[i] <> #0 DO BEGIN
    // Handle Message compression
    IF (Ord(SubLabel[i]) AND 192) = 192 THEN BEGIN
      IF Pos = 0 THEN Pos := i+2;
      i := ((Ord(SubLabel[i]) AND 63) SHL 8) + Ord(SubLabel[i+1])-PosDiff;
    END;
    Len := Ord(SubLabel[i]);
    FOR j := 1 TO Len DO Str := Str+SubLabel[i+j];
    i := i+Len+1;
    IF SubLabel[i] <> #0 THEN Str := Str+'.';
  END;
  IF Pos = 0 THEN Pos := i+1;
END;

FUNCTION TfrmMain.GetQType(qt : STRING): STRING;
BEGIN
  IF qt = QType_A_Str     THEN Result := #0+Chr(QType_A);
  IF qt = QType_NS_Str    THEN Result := #0+Chr(QType_NS);
  IF qt = QType_CNAME_Str THEN Result := #0+Chr(QType_CNAME);
  IF qt = QType_SOA_Str   THEN Result := #0+Chr(QType_SOA);
  IF qt = QType_PTR_Str   THEN Result := #0+Chr(QType_PTR);
  IF qt = QType_MX_Str    THEN Result := #0+Chr(QType_MX);
  IF qt = QType_ALL_Str   THEN Result := #0+Chr(QType_ALL);
END;

{-------------------------------------}

PROCEDURE TfrmMain.UDPClientData(Sender: TObject; Socket: TSocket);
VAR i           : WORD;
    RRDomain    : STRING;
    RRDataType  : WORD;
    uShort      : TuShort;
    DataStrings : TStringList;

    PROCEDURE GetCNAME(Buffer : TBufferRec;VAR Pos : WORD; Strings : TStringList);
    VAR DataStr : STRING;
    BEGIN
      Strings.Clear;
      LabelToString(@Buffer.Data,Pos,DataStr);
      Strings.Add(RRDomain+' CNAME '+DataStr);
    END;

    PROCEDURE GetPTR(Buffer : TBufferRec;VAR Pos : WORD; Strings : TStringList);
    VAR DataStr : STRING;
    BEGIN
      Strings.Clear;
      LabelToString(@Buffer.Data,Pos,DataStr);
      Strings.Add(RRDomain+' PTR '+DataStr);
    END;

    PROCEDURE GetNS(Buffer : TBufferRec;VAR Pos : WORD; Strings : TStringList);
    VAR DataStr : STRING;
    BEGIN
      Strings.Clear;
      LabelToString(@Buffer.Data,Pos,DataStr);
      Strings.Add(RRDomain+' NS '+DataStr);
    END;

    PROCEDURE GetA(Buffer : TBufferRec;VAR Pos : WORD; Strings : TStringList);
    VAR DataStr : STRING;
        j       : WORD;
    BEGIN
      Strings.Clear;
      DataStr := '';
      FOR j := 1 TO 3 DO BEGIN
        DataStr := DataStr+IntToStr(Buffer.nData[i])+'.';
        Inc(i);
      END;
      DataStr := DataStr+IntToStr(Buffer.nData[i]);
      Inc(i);
      Strings.Add(RRDomain+' A '+DataStr);
    END;

    PROCEDURE GetSOA(Buffer : TBufferRec;VAR Pos : WORD; Strings : TStringList);
    VAR DataStr : STRING;
        j       : WORD;
        uLong   : TuLong;
    BEGIN
      Strings.Clear;
      LabelToString(@Buffer.Data,i,DataStr);
      Strings.Add(RRDomain+' SOA '+DataStr);
      LabelToString(@Buffer.Data,i,DataStr);
      Strings.Add('Mailbox: '+DataStr);
      // Serial
      FOR j := 3 DOWNTO 0 DO BEGIN
        uLong.B[j] := Buffer.nData[i];
        Inc(i);
      END;
      Strings.Add('Serial: '+IntToStr(uLong.Value));
      // Refresh
      FOR j := 3 DOWNTO 0 DO BEGIN
        uLong.B[j] := Buffer.nData[i];
        Inc(i);
      END;
      Strings.Add('Refresh: '+IntToStr(uLong.Value));
      // Retry
      FOR j := 3 DOWNTO 0 DO BEGIN
        uLong.B[j] := Buffer.nData[i];
        Inc(i);
      END;
      Strings.Add('Retry: '+IntToStr(uLong.Value));
      // Expire
      FOR j := 3 DOWNTO 0 DO BEGIN
        uLong.B[j] := Buffer.nData[i];
        Inc(i);
      END;
      Strings.Add('Expire: '+IntToStr(uLong.Value));
      // Minimum TTL
      FOR j := 3 DOWNTO 0 DO BEGIN
        uLong.B[j] := Buffer.nData[i];
        Inc(i);
      END;
      Strings.Add('Min TTL: '+IntToStr(uLong.Value));
    END;

    PROCEDURE GetMX(Buffer : TBufferRec;VAR Pos : WORD; Strings : TStringList);
    VAR DataStr : STRING;
        Data,j  : WORD;
    BEGIN
      Strings.Clear;
      Data := Buffer.nData[i] SHL 8;
      Inc(i);
      Data := Data + Buffer.nData[i];
      Inc(i);
      DataStr := '';
      LabelToString(@Buffer.Data,i,DataStr);
      Strings.Add(RRDomain+' MX '+IntToStr(Data)+' '+DataStr);
    END;

BEGIN
  DataStrings := TStringList.Create;
  TRY
    FillChar(Buffer.ID, 524, 32);
    WITH UDPClient DO BEGIN
      UDPClient.ReadBuffer(@Buffer.ID, 524);
      TimeOut.Enabled := FALSE;
      // Check if correct Data is recieved
      IF Buffer.ID = Swap(QueryID) THEN BEGIN
        // Skip Queryrecord
        i := 0;
        WHILE Buffer.Data[i] <> #0 DO Inc(i);
        i := i+5;
        // Process Ressourcerecords
        WHILE Buffer.Data[i] <> #32 DO BEGIN
          LabelToString(@Buffer.Data,i,RRDomain);
          Inc(i);
          RRDataType := Ord(Buffer.Data[i]);
          // Skip TTL & RDLength field
          i := i+9;
          // Get Type of RR
          CASE RRDataType OF
            QType_CNAME : GetCNAME(Buffer,i,DataStrings);
            QType_NS    : GetNS(Buffer,i,DataStrings);
            QType_A     : GetA(Buffer,i,DataStrings);
            QType_SOA   : GetSOA(Buffer,i,DataStrings);
            QType_PTR   : GetPTR(Buffer,i,DataStrings);
            QType_MX    : GetMX(Buffer,i,DataStrings);
            ELSE BEGIN
              // Skip unsupported Records
              uShort.B[1] := Buffer.nData[i-2];
              uShort.B[0] := Buffer.nData[i-1];
              i := i+uShort.Value;
            END;
          END;
          TTY(DataStrings.Text);
        END;
      END ELSE TTY('Error! Recieved wrong Data from Server!');
      //
      Waiting := FALSE;
    END;
  FINALLY
    DataStrings.Free;
  END;
END;


PROCEDURE TfrmMain.btnQueryClick(Sender: TObject);
VAR SubLabel : TLabelStr;
    QType    : STRING[2];
BEGIN
  WITH UDPClient DO BEGIN
    Host := cbHost.Text;
    IF cbHost.Items.IndexOf(cbHost.Text) < 0 THEN cbHost.Items.Add(cbHost.Text);
    Port := DNSPort;
    Open;
    IF SocketState = ssOpen THEN
    BEGIN
      cbHost.Enabled := FALSE;
      edData.Enabled := FALSE;
      btnQuery.Enabled := FALSE;
      // Initialize Header Section
      Buffer.Count := 0;
      Buffer.ID := Swap(QueryID);
      Move(Header_Option[1],Buffer.Option,10);
      // Convert Datastring to Sequence of Labels
      StringToLabel(edData.Text,SubLabel);
      Move(SubLabel[1],Buffer.Data[0],Length(SubLabel));
      Buffer.Count := Length(SubLabel);
      // ADD QType and QClass
      QType := GetQType(cbType.Text);
      Move(QType[1],Buffer.Data[Buffer.Count],2);
      Buffer.Count := Buffer.Count+2;
      Move(QClass_Internet,Buffer.Data[Buffer.Count],2);
      Buffer.Count := Buffer.Count+14;  // 2 + 12
      WriteBuffer(@Buffer.ID,Buffer.Count);
      //
      TTY('Query for "'+edData.Text+'"');
      //
      TimeOut := TTimer.Create(NIL);
      TRY
        TimeOut.OnTimer := OnTimeOut;
        TimeOut.Interval := MaxTimeOut*1000;
        Waiting := TRUE;
        WHILE Waiting DO BEGIN
          Application.ProcessMessages;
          Sleep(0);
        END;
      FINALLY
        TimeOut.Free;
      END;
      //
      Close;
      cbHost.Enabled := TRUE;
      edData.Enabled := TRUE;
      btnQuery.Enabled := TRUE;
    END;
    Inc(QueryID);
  END;
END;


END.
